#include "SimpleMultiDevice.hpp"

int
Device::createContext()
{
    //Create context using current device's ID
    context = clCreateContext(cprops, 
                              1, 
                              &deviceId, 
                              0, 
                              0, 
                              &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateContext failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::createQueue()
{
    //Create Command-Queue
    queue = clCreateCommandQueue(context, 
                                 deviceId, 
                                 CL_QUEUE_PROFILING_ENABLE, 
                                 &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateCommandQueue failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::createBuffers()
{
    // Create input buffer
    inputBuffer = clCreateBuffer(context, 
                                 CL_MEM_READ_ONLY, 
                                 width * sizeof(cl_float), 
                                 0, 
                                 &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateBuffer failed.(inputBuffer)"))
        return SDK_FAILURE;

    // Create output buffer
    outputBuffer = clCreateBuffer(context, 
                                  CL_MEM_WRITE_ONLY, 
                                  width * sizeof(cl_float), 
                                  0, 
                                  &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateBuffer failed.(outputBuffer)"))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::enqueueWriteBuffer()
{
    // Initialize input buffer
    status = clEnqueueWriteBuffer(queue,
                                  inputBuffer,
                                  1, 
                                  0, 
                                  width * sizeof(cl_float),
                                  input,
                                  0, 
                                  0, 
                                  0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clEnqueueWriteBuffer failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::createProgram(const char **source, const size_t *sourceSize)
{
    // Create program with source
    program = clCreateProgramWithSource(context, 
                                        1, 
                                        source, 
                                        sourceSize, 
                                        &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateProgramWithSource failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::buildProgram()
{
    char buildOptions[50];
    sprintf(buildOptions, "-D KERNEL_ITERATIONS=%d", KERNEL_ITERATIONS);
    // Build program source
    status = clBuildProgram(program, 
                            1, 
                            &deviceId, 
                            buildOptions, 
                            0, 
                            0);
    /* Print build log here if build program failed */
    if(status != CL_SUCCESS)
    {
        if(status == CL_BUILD_PROGRAM_FAILURE)
        {
            cl_int logStatus;
            char *buildLog = NULL;
            size_t buildLogSize = 0;
            logStatus = clGetProgramBuildInfo(program, 
                                              deviceId, 
                                              CL_PROGRAM_BUILD_LOG, 
                                              buildLogSize, 
                                              buildLog, 
                                              &buildLogSize);
            if(!sdkObject.checkVal(logStatus,
                                   CL_SUCCESS,
                                   "clGetProgramBuildInfo failed."))

            {
                  return SDK_FAILURE;
            }
            
            buildLog = (char*)malloc(buildLogSize);
            if(buildLog == NULL)
            {
                sdkObject.error("Failed to allocate host memory. (buildLog)");
                return SDK_FAILURE;
            }
            memset(buildLog, 0, buildLogSize);

            logStatus = clGetProgramBuildInfo(program, 
                                              deviceId, 
                                              CL_PROGRAM_BUILD_LOG, 
                                              buildLogSize, 
                                              buildLog, 
                                              NULL);
            if(!sdkObject.checkVal(logStatus,
                                   CL_SUCCESS,
                                   "clGetProgramBuildInfo failed."))
            {
                  free(buildLog);
                  return SDK_FAILURE;
            }

            std::cout << " \n\t\t\tBUILD LOG\n";
            std::cout << " ************************************************\n";
            std::cout << buildLog << std::endl;
            std::cout << " ************************************************\n";
            free(buildLog);
        }

          if(!sdkObject.checkVal(status,
                                 CL_SUCCESS,
                                 "clBuildProgram failed."))
          {
                return SDK_FAILURE;
          }
    }

    return SDK_SUCCESS;
}

int
Device::createKernel()
{
    kernel = clCreateKernel(program, "multiDeviceKernel", &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateKernel failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::setKernelArgs()
{
    status = clSetKernelArg(kernel, 0, sizeof(cl_mem), &inputBuffer);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clSetKernelArg failed.(inputBuffer)"))
        return SDK_FAILURE;

    status = clSetKernelArg(kernel, 1, sizeof(cl_mem), &outputBuffer);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clSetKernelArg failed.(outputBuffer)"))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::enqueueKernel(size_t *globalThreads, size_t *localThreads)
{
    status = clEnqueueNDRangeKernel(queue,
                                    kernel,
                                    1,
                                    NULL,
                                    globalThreads,
                                    localThreads,
                                    0,
                                    NULL,
                                    &eventObject);
     if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clEnqueueNDRangeKernel failed."))
        return SDK_FAILURE;

    status = clFlush(queue);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clFlush failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::waitForKernel()
{
    status = clFinish(queue);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clFinish failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::getProfilingData()
{
    status = clGetEventProfilingInfo(eventObject,
                                     CL_PROFILING_COMMAND_START,
                                     sizeof(cl_ulong),
                                     &kernelStartTime,
                                     0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetEventProfilingInfo failed.(start time)"))
        return SDK_FAILURE;

    status = clGetEventProfilingInfo(eventObject,
                                     CL_PROFILING_COMMAND_END,
                                     sizeof(cl_ulong),
                                     &kernelEndTime,
                                     0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetEventProfilingInfo failed.(end time)"))
        return SDK_FAILURE;

    //Measure time in ms
    elapsedTime = 1e-6 * (kernelEndTime - kernelStartTime);

    return SDK_SUCCESS;
}

int
Device::enqueueReadData()
{
    // Allocate memory 
    if(output == NULL)
    {
        output = (cl_float*)malloc(width * sizeof(cl_float));
        if(output == NULL)
        {
            printf("Failed to allocate output buffer!\n");
            return SDK_FAILURE;
        }
    }

    status = clEnqueueReadBuffer(queue,
                                 outputBuffer,
                                 1,
                                 0,
                                 width * sizeof(cl_float),
                                 output,
                                 0, 0, 0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clEnqueueReadBuffer failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int
Device::verifyResults()
{
    float error = 0;
    //compare results between verificationOutput and output host buffers
    for(int i = 0; i < width; i++)
    {
        error += (output[i] - verificationOutput[i]);
    }
    error /= width;

    if(error < 0.001)
    {
        std::cout << "Passed!\n";
        verificationCount++;
    }
    else
    {
        std::cout << "Failed!\n";
    }

    return SDK_SUCCESS;
}

int
Device::cleanupResources()
{
    int status = clReleaseCommandQueue(queue);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseCommandQueue failed."))
        return SDK_FAILURE;

    status = clReleaseKernel(kernel);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseKernel failed."))
        return SDK_FAILURE;

    cl_uint programRefCount;
    status = clGetProgramInfo(program,
                              CL_PROGRAM_REFERENCE_COUNT,
                              sizeof(cl_uint),
                              &programRefCount,
                              0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetProgramInfo failed."))
        return SDK_FAILURE;
    if(programRefCount)
    {
        status = clReleaseProgram(program);
        if(!sdkObject.checkVal(status,
                               CL_SUCCESS,
                               "clReleaseProgram failed."))
        return SDK_FAILURE;
    }

    cl_uint inputRefCount;
    status = clGetMemObjectInfo(inputBuffer,
                                CL_MEM_REFERENCE_COUNT,
                                sizeof(cl_uint),
                                &inputRefCount,
                                0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetMemObjectInfo failed."))
        return SDK_FAILURE;
    if(inputRefCount)
    {
        status = clReleaseMemObject(inputBuffer);
        if(!sdkObject.checkVal(status,
                               CL_SUCCESS,
                               "clReleaseMemObject failed. (inputBuffer)"))
            return SDK_FAILURE;
    }

    cl_uint outputRefCount;
    status = clGetMemObjectInfo(outputBuffer,
                                CL_MEM_REFERENCE_COUNT,
                                sizeof(cl_uint),
                                &outputRefCount,
                                0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetMemObjectInfo failed."))
        return SDK_FAILURE;
    if(outputRefCount)
    {
        status = clReleaseMemObject(outputBuffer);
        if(!sdkObject.checkVal(status,
                               CL_SUCCESS,
                               "clReleaseMemObject failed. (outputBuffer)"))
            return SDK_FAILURE;
    }

    cl_uint contextRefCount;
    status = clGetContextInfo(context, 
                              CL_CONTEXT_REFERENCE_COUNT, 
                              sizeof(cl_uint), 
                              &contextRefCount, 
                              0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetContextInfo failed."))
        return SDK_FAILURE;

    if(contextRefCount)
    {
        status = clReleaseContext(context);
        if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseContext failed."))
            return SDK_FAILURE;
    }

    status = clReleaseEvent(eventObject);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseEvent failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}



//Thread function for a device
void* threadFunc(void *device)
{
    Device *d = (Device*)device;

    size_t globalThreads = width;
    size_t localThreads = GROUP_SIZE;

    d->enqueueKernel(&globalThreads, &localThreads);
    d->waitForKernel();

    return NULL;
}

Device::~Device()
{
    if(output)
    {
        free(output);
        output = NULL;
    }
}
        

int runMultiGPU()
{
    int status;

    ///////////////////////////////////////////////////////////////////
    //  Case 1 : Single Context (Single Thread)
    //////////////////////////////////////////////////////////////////
    std::cout << sep << "\nMulti GPU Test 1 : Single context Single Thread\n" << 
        sep << std::endl;

    cl_context context = clCreateContextFromType(cprops,
                                                 CL_DEVICE_TYPE_GPU,
                                                 0,
                                                 0,
                                                 &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateContext failed."))
        return SDK_FAILURE;

    size_t sourceSize = strlen(source);

    cl_program program  = clCreateProgramWithSource(context, 
                                                    1, 
                                                    &source, 
                                                    (const size_t*)&sourceSize, 
                                                    &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateProgramWithSource failed."))
        return SDK_FAILURE;

    char buildOptions[50];
    sprintf(buildOptions, "-D KERNEL_ITERATIONS=%d", KERNEL_ITERATIONS);

    //Build program for all the devices in the context
    status = clBuildProgram(program, 0, 0, buildOptions, 0, 0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clBuildProgram failed."))
        return SDK_FAILURE;


    //Setup for all GPU devices
    for(int i = 0; i < numGPUDevices; i++)
    {
        gpu[i].context = context;
        gpu[i].program = program;

        status = gpu[i].createQueue();
        if(status != SDK_SUCCESS)
            return SDK_FAILURE;

        status = gpu[i].createKernel();
        if(status != SDK_SUCCESS)
            return SDK_FAILURE;
    }

    // Create buffers
    // Create input buffer
    cl_mem inputBuffer = clCreateBuffer(context, 
                                 CL_MEM_READ_ONLY, 
                                 width * sizeof(cl_float), 
                                 0, 
                                 &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateBuffer failed.(inputBuffer)"))
        return SDK_FAILURE;


    // Create output buffer
    cl_mem outputBuffer = clCreateBuffer(context, 
                                  CL_MEM_WRITE_ONLY, 
                                  width * sizeof(cl_float), 
                                  0, 
                                  &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateBuffer failed.(outputBuffer)"))
        return SDK_FAILURE;

    //Set Buffers
    for(int i = 0; i < numGPUDevices; i++)
    {
        gpu[i].inputBuffer = inputBuffer;
        status = gpu[i].enqueueWriteBuffer();
        if(status != SDK_SUCCESS)
            return status;
        gpu[i].outputBuffer = outputBuffer;
    }


    //Set kernel arguments
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].setKernelArgs();
        if(status != SDK_SUCCESS)
            return status;
    }

    size_t globalThreads = width;
    size_t localThreads = GROUP_SIZE;

    //Start a host timer here
    int timer = sdkObject.createTimer();
    sdkObject.resetTimer(timer);
    sdkObject.startTimer(timer);

    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].enqueueKernel(&globalThreads, &localThreads);
        if(status != SDK_SUCCESS)
            return status;
    }

    //Wait for all kernels to finish execution
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].waitForKernel();
        if(status != SDK_SUCCESS)
            return status;
    }

    //Stop the host timer here
    sdkObject.stopTimer(timer);

    //Measure total time
    double totalTime = sdkObject.readTimer(timer);

    //Get individual timers
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].getProfilingData();
        if(status != SDK_SUCCESS)
            return status;
    }


    //Print total time and individual times
    std::cout << "Total time : " << totalTime * 1000 << std::endl;
    for(int i = 0; i < numGPUDevices; i++)
    {
        std::cout << "Time of GPU" << i << " : " << gpu[i].elapsedTime << 
            std::endl;
    }

    if(verify)
    {
        //Enqueue Read output buffer and verify results
        for(int i = 0; i < numGPUDevices; i++)
        {
            status = gpu[i].enqueueReadData();
            if(status != SDK_SUCCESS)
                return status;

            // Verify results
            std::cout << "Verifying results for GPU" << i << " : ";

            gpu[i].verifyResults();
        }
    }

    //Release the resources on all devices
    //Release context
    status = clReleaseContext(context);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateContext failed."))
        return SDK_FAILURE;


    //Release memory buffers
    status = clReleaseMemObject(inputBuffer);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseMemObject failed. (inputBuffer)"))
        return SDK_FAILURE;


    status = clReleaseMemObject(outputBuffer);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseMemObject failed. (outputBuffer)"))
        return SDK_FAILURE;



    //Release Program object
    status = clReleaseProgram(program);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseProgram failed."))
        return SDK_FAILURE;

    //Release Kernel object, command-queue, event object
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = clReleaseKernel(gpu[i].kernel);
        if(!sdkObject.checkVal(status,
                               CL_SUCCESS,
                               "clReleaseCommandQueue failed."))
            return SDK_FAILURE;

        status = clReleaseCommandQueue(gpu[i].queue);
        if(!sdkObject.checkVal(status,
                               CL_SUCCESS,
                               "clReleaseCommandQueue failed."))
            return SDK_FAILURE;

        status = clReleaseEvent(gpu[i].eventObject);
        if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseEvent failed."))
        return SDK_FAILURE;
    }



    ///////////////////////////////////////////////////////////////////
    //  Case 2 : Multiple Context (Single Thread)
    //////////////////////////////////////////////////////////////////
    std::cout << sep << "\nMulti GPU Test 2 : Multiple context Single Thread\n" <<
        sep << std::endl;

    for(int i = 0; i < numGPUDevices; i++)
    {
        //Create context for each device
        status = gpu[i].createContext();
        if(status != SDK_SUCCESS)
            return status;

        //Create command-queue;
        status = gpu[i].createQueue();
        if(status != SDK_SUCCESS)
            return status;

        //Create memory buffers
        status = gpu[i].createBuffers();
        if(status != SDK_SUCCESS)
            return status;

        //Initialize input buffer
        status = gpu[i].enqueueWriteBuffer();
        if(status != SDK_SUCCESS)
            return status;

        //create program object
        status = gpu[i].createProgram(&source, &sourceSize);
        if(status != SDK_SUCCESS)
            return status;

        //Build program
        status = gpu[i].buildProgram();
        if(status != SDK_SUCCESS)
            return status;

        //Create kernel objects for each device
        status = gpu[i].createKernel();
        if(status != SDK_SUCCESS)
            return status;
    }
    
    //Set kernel arguments
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].setKernelArgs();
        if(status != SDK_SUCCESS)
            return status;
    }

    //Start a host timer here
    sdkObject.resetTimer(timer);
    sdkObject.startTimer(timer);

    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].enqueueKernel(&globalThreads, &localThreads);
        if(status != SDK_SUCCESS)
            return status;
    }

    //Wait for all kernels to finish execution
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].waitForKernel();
        if(status != SDK_SUCCESS)
            return status;
    }

    //Stop the host timer here
    sdkObject.stopTimer(timer);

    //Measure total time
    totalTime = sdkObject.readTimer(timer);

    //Get individual timers
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].getProfilingData();
        if(status != SDK_SUCCESS)
            return status;
    }

    //Print total time and individual times
    std::cout << "Total time : " << totalTime * 1000 << std::endl;
    for(int i = 0; i < numGPUDevices; i++)
    {
        std::cout << "Time of GPU" << i << " : " << gpu[i].elapsedTime << 
            std::endl;
    }

    if(verify)
    {
        // Read outputdata and verify results
        for(int i = 0; i < numGPUDevices; i++)
        {
            status = gpu[i].enqueueReadData();
            if(status != SDK_SUCCESS)
                return status;

            // Verify results
            std::cout << "Verifying results for GPU" << i << " : ";

            gpu[i].verifyResults();
        }
    }


    //Release the resources on all devices
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].cleanupResources();
        if(status != SDK_SUCCESS)
            return status;
    }

    ////////////////////////////////////////////////////////////////////
    //  Case 3 : Multiple thread and multiple context for each device
    ////////////////////////////////////////////////////////////////////
    std::cout << sep << "\nMulti GPU Test 3 : Multiple context Multiple Thread\n" <<
        sep << std::endl;

    for(int i = 0; i < numGPUDevices; i++)
    {
        //Create context for each device
        status = gpu[i].createContext();
        if(status != SDK_SUCCESS)
            return status;

        //Create command-queue;
        status = gpu[i].createQueue();
        if(status != SDK_SUCCESS)
            return status;

        //Create memory buffers
        status = gpu[i].createBuffers();
        if(status != SDK_SUCCESS)
            return status;

        //Initialize input buffer
        status = gpu[i].enqueueWriteBuffer();
        if(status != SDK_SUCCESS)
            return status;

        //create program object
        status = gpu[i].createProgram(&source, &sourceSize);
        if(status != SDK_SUCCESS)
            return status;

        //Build program
        status = gpu[i].buildProgram();
        if(status != SDK_SUCCESS)
            return status;

        //Create kernel objects for each device
        status = gpu[i].createKernel();
        if(status != SDK_SUCCESS)
            return status;
    }
    
    //Set kernel arguments
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].setKernelArgs();
        if(status != SDK_SUCCESS)
            return status;
    }

    //Start a host timer here
    sdkObject.resetTimer(timer);
    sdkObject.startTimer(timer);

    //Create thread objects
    streamsdk::SDKThread *gpuThread = new streamsdk::SDKThread[numGPUDevices];

    //Start threads for each gpu device
    for(int i = 0; i < numGPUDevices; i++)
    {
        gpuThread[i].create(threadFunc, (void *)(gpu + i));
    }

    //Join all gpu threads
    for(int i = 0; i < numGPUDevices; i++)
    {
        gpuThread[i].join();
    }

    delete []gpuThread;

    //Stop the host timer here
    sdkObject.stopTimer(timer);

    //Measure total time
    totalTime = sdkObject.readTimer(timer);

    //Get individual timers
    for(int i = 0; i < numGPUDevices; i++)
    {
        status = gpu[i].getProfilingData();
        if(status != SDK_SUCCESS)
            return status;
    }

    //Print total time and individual times
    std::cout << "Total time : " << totalTime * 1000 << std::endl;
    for(int i = 0; i < numGPUDevices; i++)
    {
        std::cout << "Time of GPU" << i << " : " << gpu[i].elapsedTime << 
            std::endl;
    }

    if(verify)
    {
        // Read outputdata and verify results
        for(int i = 0; i < numGPUDevices; i++)
        {
            status = gpu[i].enqueueReadData();
            if(status != SDK_SUCCESS)
                return status;

            // Verify results
            std::cout << "Verifying results for GPU" << i << " : ";

            gpu[i].verifyResults();
        }
    }

    //Release the resources on all devices
    for(int i = 0; i < numGPUDevices; i++)
    {
        gpu[i].cleanupResources();
    }
    return 0;
}

int runMultiDevice()
{
    int status;

    ///////////////////////////////////////////////////////////////////
    //  Case 1 : Single Context (Single Thread)
    //////////////////////////////////////////////////////////////////
    std::cout << sep << "\nCPU + GPU Test 1 : Single context Single Thread\n" << 
        sep << std::endl;

    /* Create a list of device IDs having only CPU0 and GPU0 as device IDs */
    cl_device_id *devices = (cl_device_id*)malloc(2 * sizeof(cl_device_id));
    devices[0] = cpu[0].deviceId;
    devices[1] = gpu[0].deviceId;

    cl_context context = clCreateContext(cprops, 2, devices, 0, 0, &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateContext failed."))
        return SDK_FAILURE;

    size_t sourceSize = strlen(source);

    cl_program program  = clCreateProgramWithSource(context, 
                                                    1, 
                                                    &source, 
                                                    (const size_t*)&sourceSize, 
                                                    &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateProgramWithSource failed."))
        return SDK_FAILURE;

    char buildOptions[50];
    sprintf(buildOptions, "-D KERNEL_ITERATIONS=%d", KERNEL_ITERATIONS);

    //Build program for all the devices in the context
    status = clBuildProgram(program, 0, 0, buildOptions, 0, 0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clBuildProgram failed."))
        return SDK_FAILURE;

    //Allocate objects for CPU
    cpu[0].context = context;
    gpu[0].context = context;

    cpu[0].program = program;
    gpu[0].program = program;

    // Create command queue 
    status = cpu[0].createQueue();
    if(status != SDK_SUCCESS)
        return status;

    // Create kernel
    status = cpu[0].createKernel();
    if(status != SDK_SUCCESS)
        return status;

    // Create queue
    status = gpu[0].createQueue();
    if(status != SDK_SUCCESS)
        return status;

    // Create kernel
    status = gpu[0].createKernel();
    if(status != SDK_SUCCESS)
        return status;

    // Create buffers - A buffer is created on all devices sharing a context
    // So bufffer creation should should not per device in a single-context
    cl_mem inputBuffer = clCreateBuffer(context, 
                                        CL_MEM_READ_ONLY, 
                                        width * sizeof(cl_float), 
                                        0, 
                                        &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateBuffer failed.(inputBuffer)"))
        return SDK_FAILURE;

    cl_mem outputBuffer = clCreateBuffer(context, 
                                         CL_MEM_WRITE_ONLY, 
                                         width * sizeof(cl_float), 
                                         0, 
                                         &status);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateBuffer failed.(outputBuffer)"))
        return SDK_FAILURE;

    cpu[0].inputBuffer = inputBuffer;
    gpu[0].inputBuffer = inputBuffer;

    cpu[0].outputBuffer = outputBuffer;
    gpu[0].outputBuffer = outputBuffer;

    // Initialize input buffer for both devices
    status = cpu[0].enqueueWriteBuffer();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].enqueueWriteBuffer();
    if(status != SDK_SUCCESS)
        return status;

    //Set kernel arguments
    status = cpu[0].setKernelArgs();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].setKernelArgs();
    if(status != SDK_SUCCESS)
        return status;


    size_t globalThreads = width;
    size_t localThreads = GROUP_SIZE;

    //Start a host timer here
    int timer = sdkObject.createTimer();
    sdkObject.resetTimer(timer);
    sdkObject.startTimer(timer);

    status = cpu[0].enqueueKernel(&globalThreads, &localThreads);
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].enqueueKernel(&globalThreads, &localThreads);
    if(status != SDK_SUCCESS)
        return status;

    //Wait for all kernels to finish execution
    status = cpu[0].waitForKernel();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].waitForKernel();
    if(status != SDK_SUCCESS)
        return status;

    //Stop the host timer here
    sdkObject.stopTimer(timer);

    //Measure total time
    double totalTime = sdkObject.readTimer(timer);

    status = cpu[0].getProfilingData();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].getProfilingData();
    if(status != SDK_SUCCESS)
        return status;

    //Print total time and individual times
    std::cout << "Total time : " << totalTime * 1000 << std::endl;
    std::cout << "Time of CPU : " << cpu[0].elapsedTime << std::endl;
    std::cout << "Time of GPU : " << gpu[0].elapsedTime << std::endl;

   
    if(verify)
    {
        //Read back output data for verification
        status = cpu[0].enqueueReadData();
        if(status != SDK_SUCCESS)
            return status;

        status = gpu[0].enqueueReadData();
        if(status != SDK_SUCCESS)
            return status;

        // Verify results
        std::cout << "Verifying results for CPU : ";
        cpu[0].verifyResults();
        std::cout << "Verifying results for GPU : ";
        gpu[0].verifyResults();
    }

    //Release context
    status = clReleaseContext(context);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clCreateContext failed."))
        return SDK_FAILURE;


    //Release memory buffers
    status = clReleaseMemObject(inputBuffer);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseMemObject failed. (inputBuffer)"))
        return SDK_FAILURE;


    status = clReleaseMemObject(outputBuffer);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseMemObject failed. (outputBuffer)"))
        return SDK_FAILURE;

    //ReleaseCommand-queue
    status = clReleaseCommandQueue(cpu[0].queue);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseCommandQueue failed."))
        return SDK_FAILURE;

    status = clReleaseCommandQueue(gpu[0].queue);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseCommandQueue failed."))
        return SDK_FAILURE;

    //Release Program object
    status = clReleaseProgram(program);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseProgram failed."))
        return SDK_FAILURE;

    //Release Kernel object
    status = clReleaseKernel(cpu[0].kernel);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseKernel failed."))
        return SDK_FAILURE;

    status = clReleaseKernel(gpu[0].kernel);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseKernel failed."))
        return SDK_FAILURE;

    //Release Event object
    status = clReleaseEvent(cpu[0].eventObject);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseEvent failed."))
        return SDK_FAILURE;

    status = clReleaseEvent(gpu[0].eventObject);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clReleaseEvent failed."))
        return SDK_FAILURE;


    //Release the resources on all devices
    /*status = cpu[0].cleanupResources();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].cleanupResources();
    if(status != SDK_SUCCESS)
        return status;*/


    ///////////////////////////////////////////////////////////////////
    //  Case 2 : Multiple Context (Single Thread)
    //////////////////////////////////////////////////////////////////
    std::cout << sep << "\nCPU + GPU Test 2 : Multiple context Single Thread\n" <<
        sep << std::endl;

    status = cpu[0].createContext();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createQueue();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createBuffers();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createProgram(&source, &sourceSize);
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].buildProgram();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createKernel();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createContext();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createQueue();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createBuffers();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createProgram(&source, &sourceSize);
    if(status != SDK_SUCCESS)
        return status;
    
    status = gpu[0].buildProgram();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createKernel();
    if(status != SDK_SUCCESS)
        return status;

    // Initialize input buffer for both devices
    status = cpu[0].enqueueWriteBuffer();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].enqueueWriteBuffer();
    if(status != SDK_SUCCESS)
        return status;
    
    //Set kernel arguments
    status = cpu[0].setKernelArgs();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].setKernelArgs();
    if(status != SDK_SUCCESS)
        return status;

    //Start a host timer here
    //int timer = sdkObject.createTimer();
    sdkObject.resetTimer(timer);
    sdkObject.startTimer(timer);

    //size_t globalThreads = width;
    //size_t localThreads = 1;
    status = cpu[0].enqueueKernel(&globalThreads, &localThreads);
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].enqueueKernel(&globalThreads, &localThreads);
    if(status != SDK_SUCCESS)
        return status;

    //Wait for all kernels to finish execution
    status = cpu[0].waitForKernel();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].waitForKernel();
    if(status != SDK_SUCCESS)
        return status;

    //Stop the host timer here
    sdkObject.stopTimer(timer);

    //Measure total time
    //double totalTime = sdkObject.readTimer(timer);

    status = cpu[0].getProfilingData();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].getProfilingData();
    if(status != SDK_SUCCESS)
        return status;

    //Print total time and individual times
    std::cout << "Total time : " << totalTime * 1000 << std::endl;
    std::cout << "Time of CPU : " << cpu[0].elapsedTime << std::endl;
    std::cout << "Time of GPU : " << gpu[0].elapsedTime << std::endl;

   
    if(verify)
    {
        //Read back output data for verification
        status = cpu[0].enqueueReadData();
        if(status != SDK_SUCCESS)
            return status;

        status = gpu[0].enqueueReadData();
        if(status != SDK_SUCCESS)
            return status;

        // Verify results
        std::cout << "Verifying results for CPU : ";
        cpu[0].verifyResults();
        std::cout << "Verifying results for GPU : ";
        gpu[0].verifyResults();
    }

    //Release the resources on all devices
    status = cpu[0].cleanupResources();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].cleanupResources();
    if(status != SDK_SUCCESS)
        return status;

    /////////////////////////////////////////////////////////////////////
    //  Case 3 : Multiple thread and multiple context for each device
    ////////////////////////////////////////////////////////////////////
    std::cout << sep << "\nCPU + GPU Test 3 : Multiple context Multiple Thread\n" <<
        sep << std::endl;

    status = cpu[0].createContext();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createQueue();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createBuffers();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createProgram(&source, &sourceSize);
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].buildProgram();
    if(status != SDK_SUCCESS)
        return status;

    status = cpu[0].createKernel();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createContext();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createQueue();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createBuffers();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createProgram(&source, &sourceSize);
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].buildProgram();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].createKernel();
    if(status != SDK_SUCCESS)
        return status;

    // Initialize input buffer for both devices
    status = cpu[0].enqueueWriteBuffer();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].enqueueWriteBuffer();
    if(status != SDK_SUCCESS)
        return status;

    //Set kernel arguments
    status = cpu[0].setKernelArgs();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].setKernelArgs();
    if(status != SDK_SUCCESS)
        return status;

    //Start a host timer here
    sdkObject.resetTimer(timer);
    sdkObject.startTimer(timer);

    //Create a thread for CPU and GPU device each
    streamsdk::SDKThread cpuThread;
    streamsdk::SDKThread gpuThread;

    cpuThread.create(threadFunc, (void *)cpu);
    gpuThread.create(threadFunc, (void *)gpu);

    cpuThread.join();
    gpuThread.join();

    //Stop the host timer here
    sdkObject.stopTimer(timer);

    //Measure total time
    totalTime = sdkObject.readTimer(timer);

    status = cpu[0].getProfilingData();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].getProfilingData();
    if(status != SDK_SUCCESS)
        return status;

    //Print total time and individual times
    std::cout << "Total time : " << totalTime * 1000 << std::endl;
    std::cout << "Time of CPU : " << cpu[0].elapsedTime << std::endl;
    std::cout << "Time of GPU : " << gpu[0].elapsedTime << std::endl;

    if(verify)
    {
         //Read back output data for verification
        status = cpu[0].enqueueReadData();
        if(status != SDK_SUCCESS)
            return status;

        status = gpu[0].enqueueReadData();
        if(status != SDK_SUCCESS)
            return status;

        // Verify results
        std::cout << "Verifying results for CPU : ";
        cpu[0].verifyResults();
        std::cout << "Verifying results for GPU : ";
        gpu[0].verifyResults();
    }

    //Release the resources on all devices
    status = cpu[0].cleanupResources();
    if(status != SDK_SUCCESS)
        return status;

    status = gpu[0].cleanupResources();
    if(status != SDK_SUCCESS)
        return status;

    if(devices)
    {
        free(devices);
        devices = NULL;
    }

    return 0;
}
/*
 * \brief Host Initialization 
 *        Allocate and initialize memory 
 *        on the host. Print input array. 
 */
int
initializeHost(void)
{
    width = NUM_THREADS;
    input = NULL;
    verificationOutput = NULL;

    /////////////////////////////////////////////////////////////////
    // Allocate and initialize memory used by host 
    /////////////////////////////////////////////////////////////////
    cl_uint sizeInBytes = width * sizeof(cl_uint);
    input = (cl_float*) malloc(sizeInBytes);
    if(input == NULL)
    {
        printf("Error: Failed to allocate input memory on host\n");
        return SDK_FAILURE; 
    }

    verificationOutput = (cl_float*) malloc(sizeInBytes);
    if(verificationOutput == NULL)
    {
        printf("Error: Failed to allocate verificationOutput memory on host\n");
        return SDK_FAILURE; 
    }

    //Initilize input data
    for(int i = 0; i < width; i++)
        input[i] = (cl_float)i;

    return SDK_SUCCESS;
}

/*
 * Converts the contents of a file into a string
 */
std::string
convertToString(const char *filename)
{
    size_t size;
    char*  str;
    std::string s;

    std::fstream f(filename, (std::fstream::in | std::fstream::binary));

    if(f.is_open())
    {
        size_t fileSize;
        f.seekg(0, std::fstream::end);
        size = fileSize = (size_t)f.tellg();
        f.seekg(0, std::fstream::beg);

        str = new char[size+1];
        if(!str)
        {
            f.close();
            return NULL;
        }

        f.read(str, fileSize);
        f.close();
        str[size] = '\0';
    
        s = str;
        delete[] str;
        return s;
    }
    return NULL;
}


/*
 * \brief OpenCL related initialization 
 *        Create Context, Device list, Command Queue
 *        Create OpenCL memory buffer objects
 *        Load CL file, compile, link CL source 
 *		  Build program and kernel objects
 */
int
initializeCL(void)
{
    cl_int status = 0;
    /*
     * Have a look at the available platforms and pick either
     * the AMD one if available or a reasonable default.
     */

    cl_uint numPlatforms;
    status = clGetPlatformIDs(0, NULL, &numPlatforms);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetPlatformIDs failed."))
        return SDK_FAILURE;
    
    if(numPlatforms > 0)
    {
        cl_platform_id* platforms = (cl_platform_id *)malloc(numPlatforms*sizeof(cl_platform_id));
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetPlatformIDs failed."))
            return SDK_FAILURE;

        for(unsigned int i=0; i < numPlatforms; ++i)
        {
            char pbuff[100];
            status = clGetPlatformInfo(
                        platforms[i],
                        CL_PLATFORM_VENDOR,
                        sizeof(pbuff),
                        pbuff,
                        NULL);
            if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetPlatformInfo failed."))
                return SDK_FAILURE;

            platform = platforms[i];
            if(!strcmp(pbuff, "Advanced Micro Devices, Inc."))
            {
                break;
            }
        }
        free(platforms);
    }

    /* 
     * If we could find our platform, use it. Otherwise pass a NULL and get whatever the
     * implementation thinks we should be using.
     */
    cps[0] = CL_CONTEXT_PLATFORM;
    cps[1] = (cl_context_properties)platform;
    cps[2] = 0;

    cprops = (NULL == platform) ? NULL : cps;

    // Get Number of CPU devices available
    status = clGetDeviceIDs(platform,
                            CL_DEVICE_TYPE_CPU, 
                            0, 
                            0, 
                            (cl_uint*)&numCPUDevices);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetDeviceIDs failed.(numCPUDevices)"))
        return SDK_FAILURE;

    // Get Number of CPU devices available
    status = clGetDeviceIDs(platform,
                            CL_DEVICE_TYPE_ALL, 
                            0, 
                            0, 
                            (cl_uint*)&numDevices);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetDeviceIDs failed.(numDevices)"))
        return SDK_FAILURE;

    // Get number of GPU Devices
    numGPUDevices = numDevices - numCPUDevices;

    // If no GPU is present then exit
    if(numGPUDevices < 1)
    {
        sdkObject.expectedError("Only CPU device is present. Exiting!");
        return SDK_EXPECTED_FAILURE;
    }

    // Allocate memory for list of Devices
    cpu = new Device[numCPUDevices];
    //Get CPU Device IDs
    cl_device_id* cpuDeviceIDs = new cl_device_id[numCPUDevices];
    status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, numCPUDevices, cpuDeviceIDs, 0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetDeviceIDs failed."))
        return SDK_FAILURE;
    
    for(int i = 0; i < numCPUDevices; i++)
    {
        cpu[i].dType = CL_DEVICE_TYPE_CPU;
        cpu[i].deviceId = cpuDeviceIDs[i];
    }

    delete[] cpuDeviceIDs;

    gpu = new Device[numGPUDevices];
    //Get GPU Device IDs
    cl_device_id* gpuDeviceIDs = new cl_device_id[numGPUDevices];
    status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, numGPUDevices, gpuDeviceIDs, 0);
    if(!sdkObject.checkVal(status,
                           CL_SUCCESS,
                           "clGetDeviceIDs failed."))
        return SDK_FAILURE;
    
    for(int i = 0; i < numGPUDevices; i++)
    {
        gpu[i].dType = CL_DEVICE_TYPE_GPU;
        gpu[i].deviceId = gpuDeviceIDs[i];
    }

    delete[] gpuDeviceIDs;

    /////////////////////////////////////////////////////////////////
    // Load CL file
    /////////////////////////////////////////////////////////////////
    const char *filename  = "SimpleMultiDevice_Kernels.cl";
    sourceStr = convertToString(filename);
    source = sourceStr.c_str();

    return SDK_SUCCESS;
}

int
run()
{
    int status;

    // If a GPU is present then run CPU + GPU concurrently
    if(numGPUDevices > 0 && numCPUDevices > 0)
    {
        // 3 tests : 
        // a) Single context - Single thread
        // b) Multiple context - Single thread
        // c) Multiple context - Multple Threads

        // 3 Tests * 2 devices
        requiredCount += 3 * 2;
        status = runMultiDevice();
        if(status != SDK_SUCCESS)
            return status;
    }
    
    // If more than 1 GPU is present then run MultiGPU concurrently
    if(numGPUDevices > 1)
    {
        // 3 tests : 
        // a) Single context - Single thread
        // b) Multiple context - Single thread
        // c) Multiple context - Multple Threads

        // 3 Tests * numGPUDevices
        requiredCount += 3 * numGPUDevices;
        status = runMultiGPU();
        if(status != SDK_SUCCESS)
            return status;
    }
    return SDK_SUCCESS;

}


/* 
 * \brief Releases program's resources 
 */
void
cleanupHost(void)
{
    if(input != NULL)
    {
        free(input);
        input = NULL;
    }
    if(verificationOutput != NULL)
    {
        free(verificationOutput);
        verificationOutput = NULL;
    }
    if(cpu != NULL)
    {
        delete[] cpu;
        cpu = NULL;
    }
    if(gpu != NULL)
    {
        delete[] gpu;
        gpu = NULL;
    }
}


/*
 * \brief Print no more than 256 elements of the given array.
 *
 *        Print Array name followed by elements.
 */
void print1DArray(
         const std::string arrayName, 
         const unsigned int * arrayData, 
         const unsigned int length)
{
    cl_uint i;
    cl_uint numElementsToPrint = (256 < length) ? 256 : length;

    std::cout << std::endl;
    std::cout << arrayName << ":" << std::endl;
    for(i = 0; i < numElementsToPrint; ++i)
    {
        std::cout << arrayData[i] << " ";
    }
    std::cout << std::endl;

}

// OpenCL MAD definition for CPU
float mad(float a, float b, float c)
{
    return a * b + c;
}

// OpenCL HYPOT definition for CPU
float hypot(float a, float b)
{
    return sqrt(a * a + b * b);
}


int
CPUKernel()
{
    for(int i = 0; i < width; i++)
    {
        float a = mad(input[i], input[i], 1);
        float b = mad(input[i], input[i], 2);
        
        for(int j = 0; j < KERNEL_ITERATIONS; j++)
        {
            a = hypot(a, b);
            b = hypot(a, b);
        }
        verificationOutput[i] = (a + b);
    }
    return 0;
}

int 
main(int argc, char * argv[])
{
    for(int i = 1; i < argc; i++)
    {
        if(!strcmp(argv[i], "-e") || !strcmp(argv[i], "--verify"))
            verify = true;
        if(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
        {
            printf("Usage\n");
            printf("-h, --help\tPrint this help.\n");
            printf("-e, --verify\tVerify results against reference implementation.\n");
            exit(0);
        }
    }

    int status;

    // Initialize Host application 
    status = initializeHost();
    if(status != SDK_SUCCESS)
        return status;

    // Run host computation if verification is true
    CPUKernel();

    // Initialize OpenCL resources
    status = initializeCL();
    if(status != SDK_SUCCESS)
    {
        if(status == SDK_EXPECTED_FAILURE)
            return SDK_SUCCESS;
        else
            return status;
    }

    // Run the CL program
    status = run();
    if(status != SDK_SUCCESS)
        return status;

    // Release host resources
    cleanupHost();

    if(verify)
    {
        // If any one test fails then print FAILED
        if(verificationCount != requiredCount)
        {
            std::cout << "\n\nFAILED!\n";
            return SDK_FAILURE;
        }
        else
        {
            std::cout << "\n\nPASSED!\n";
            return SDK_SUCCESS;
        }
    }

    return SDK_SUCCESS;
}
